"""OpenAI chat wrapper."""
from __future__ import annotations

import logging
import sys
from typing import (
    TYPE_CHECKING,
    Any,
    AsyncIterator,
    Callable,
    Dict,
    Iterator,
    List,
    Mapping,
    Optional,
    Tuple,
    Union,
)

from langchain.adapters.openai import convert_dict_to_message, convert_message_to_dict
from langchain.callbacks.manager import (
    AsyncCallbackManagerForLLMRun,
    CallbackManagerForLLMRun,
)
from langchain.chat_models.base import BaseChatModel
from langchain.llms.base import create_base_retry_decorator
from langchain.pydantic_v1 import Field, root_validator
from langchain.schema import ChatGeneration, ChatResult
from langchain.schema.messages import (
    AIMessageChunk,
    BaseMessage,
    BaseMessageChunk,
    ChatMessageChunk,
    FunctionMessageChunk,
    HumanMessageChunk,
    SystemMessageChunk,
)
from langchain.schema.output import ChatGenerationChunk
from langchain.utils import get_from_dict_or_env, get_pydantic_field_names

if TYPE_CHECKING:
    import tiktoken

logger = logging.getLogger(__name__)


def _import_tiktoken() -> Any:
    try:
        import tiktoken
    except ImportError:
        raise ValueError(
            "Could not import tiktoken python package. "
            "This is needed in order to calculate get_token_ids. "
            "Please install it with `pip install tiktoken`."
        )
    return tiktoken


def _create_retry_decorator(
    llm: SAPChatOpenAI,
    run_manager: Optional[
        Union[AsyncCallbackManagerForLLMRun, CallbackManagerForLLMRun]
    ] = None,
) -> Callable[[Any], Any]:
    import openai

    errors = [
        openai.error.Timeout,
        openai.error.APIError,
        openai.error.APIConnectionError,
        openai.error.RateLimitError,
        openai.error.ServiceUnavailableError,
    ]
    return create_base_retry_decorator(
        error_types=errors, max_retries=llm.max_retries, run_manager=run_manager
    )


async def acompletion_with_retry(
    llm: SAPChatOpenAI,
    run_manager: Optional[AsyncCallbackManagerForLLMRun] = None,
    **kwargs: Any,
) -> Any:
    """Use tenacity to retry the async completion call."""
    retry_decorator = _create_retry_decorator(llm, run_manager=run_manager)

    @retry_decorator
    async def _completion_with_retry(**kwargs: Any) -> Any:
        # Use OpenAI's async api https://github.com/openai/openai-python#async-api
        return await llm.client.acreate(**kwargs)

    return await _completion_with_retry(**kwargs)


def _convert_delta_to_message_chunk(
    _dict: Mapping[str, Any], default_class: type[BaseMessageChunk]
) -> BaseMessageChunk:
    role = _dict.get("role")
    content = _dict.get("content") or ""
    if _dict.get("function_call"):
        additional_kwargs = {"function_call": dict(_dict["function_call"])}
    else:
        additional_kwargs = {}

    if role == "user" or default_class == HumanMessageChunk:
        return HumanMessageChunk(content=content)
    elif role == "assistant" or default_class == AIMessageChunk:
        return AIMessageChunk(content=content, additional_kwargs=additional_kwargs)
    elif role == "system" or default_class == SystemMessageChunk:
        return SystemMessageChunk(content=content)
    elif role == "function" or default_class == FunctionMessageChunk:
        return FunctionMessageChunk(content=content, name=_dict["name"])
    elif role or default_class == ChatMessageChunk:
        return ChatMessageChunk(content=content, role=role)
    else:
        return default_class(content=content)


class SAPChatOpenAI(BaseChatModel):
    """`OpenAI` Chat large language models API.

    To use, you should have the ``openai`` python package installed, and the
    environment variable ``OPENAI_API_KEY`` set with your API key.

    Any parameters that are valid to be passed to the openai.create call can be passed
    in, even if not explicitly saved on this class.

    Example:
        .. code-block:: python

            from langchain.chat_models import ChatOpenAI
            openai = ChatOpenAI(model_name="gpt-3.5-turbo")
    """

    @property
    def lc_secrets(self) -> Dict[str, str]:
        return {"openai_api_key": "OPENAI_API_KEY"}

    @property
    def lc_serializable(self) -> bool:
        return True

    client: Any = None  #: :meta private:
    model_name: str = Field(default="gpt-3.5-turbo", alias="model")
    """Model name to use."""
    temperature: float = 0.7
    """What sampling temperature to use."""
    model_kwargs: Dict[str, Any] = Field(default_factory=dict)
    """Holds any model parameters valid for `create` call not explicitly specified."""
    openai_api_key: Optional[str] = None
    """Base URL path for API requests, 
    leave blank if not using a proxy or service emulator."""
    openai_api_base: Optional[str] = None
    openai_organization: Optional[str] = None
    """ SAP specific authentication data."""
    openai_api_client_id: Optional[str] = None
    openai_api_client_secret: Optional[str] = None
    openai_api_tokenurl: Optional[str] = None
    openai_api_url: Optional[str] = None
    # to support explicit proxy for OpenAI
    openai_proxy: Optional[str] = None
    request_timeout: Optional[Union[float, Tuple[float, float]]] = None
    """Timeout for requests to OpenAI completion API. Default is 600 seconds."""
    max_retries: int = 6
    """Maximum number of retries to make when generating."""
    streaming: bool = False
    """Whether to stream the results or not."""
    n: int = 1
    """Number of chat completions to generate for each prompt."""
    max_tokens: Optional[int] = None
    """Maximum number of tokens to generate."""
    tiktoken_model_name: Optional[str] = None
    """The model name to pass to tiktoken when using this class. 
    Tiktoken is used to count the number of tokens in documents to constrain 
    them to be under a certain limit. By default, when set to None, this will 
    be the same as the embedding model name. However, there are some cases 
    where you may want to use this Embedding class with a model name not 
    supported by tiktoken. This can include when using Azure embeddings or 
    when using one of the many model providers that expose an OpenAI-like 
    API but with different models. In those cases, in order to avoid erroring 
    when tiktoken is called, you can specify a model name to use here."""

    class Config:
        """Configuration for this pydantic object."""

        allow_population_by_field_name = True

    @root_validator(pre=True)
    def build_extra(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        """Build extra kwargs from additional params that were passed in."""
        all_required_field_names = get_pydantic_field_names(cls)
        extra = values.get("model_kwargs", {})
        for field_name in list(values):
            if field_name in extra:
                raise ValueError(f"Found {field_name} supplied twice.")
            if field_name not in all_required_field_names:
                logger.warning(
                    f"""WARNING! {field_name} is not default parameter.
                    {field_name} was transferred to model_kwargs.
                    Please confirm that {field_name} is what you intended."""
                )
                extra[field_name] = values.pop(field_name)

        invalid_model_kwargs = all_required_field_names.intersection(extra.keys())
        if invalid_model_kwargs:
            raise ValueError(
                f"Parameters {invalid_model_kwargs} should be specified explicitly. "
                f"Instead they were passed in as part of `model_kwargs` parameter."
            )

        values["model_kwargs"] = extra
        return values

    @root_validator()
    def validate_environment(cls, values: Dict) -> Dict:
        """Validate that the SAP openAI credentials and url are set"""
        values["openai_api_client_id"] = get_from_dict_or_env(
            values,
            "openai_api_client_id", 
            "OPENAI_CLIENTID"
        )
        values["openai_api_client_secret"] = get_from_dict_or_env(
            values,
            "openai_api_client_secret",
            "OPENAI_CLIENTSECRET",
            default="",
        )
        values["openai_api_url"] = get_from_dict_or_env(
            values,
            "openai_api_url",
            "OPENAI_APIURL",
            default="",
        )
        values["openai_api_tokenurl"] = get_from_dict_or_env(
            values,
            "openai_api_tokenurl",
            "OPENAI_TOKENURL",
            default="",
        )
        values["openai_proxy"] = get_from_dict_or_env(
            values,
            "openai_proxy",
            "OPENAI_PROXY",
            default="",
        )
        try:
            import saplangchainproxy.sapwrapper as sapwrapper
        except ImportError:
            raise ValueError(
                "Could not import sapopenaiwrapper python package. "
                "Check path of you project directory."
            )     
        try:
            values["client"] = sapwrapper.ChatCompletion          
        except AttributeError:
            raise ValueError(
                "`openai` has no `ChatCompletion` attribute, this is likely "
                "due to an old version of the openai package. Try upgrading it "
                "with `pip install --upgrade openai`."
            )
        if values["n"] < 1:
            raise ValueError("n must be at least 1.")
        if values["n"] > 1 and values["streaming"]:
            raise ValueError("n must be 1 when streaming.")
        return values
    
    @property
    def _default_params(self) -> Dict[str, Any]:
        """Get the default parameters for calling OpenAI API."""
        return {
            "model": self.model_name,
            "request_timeout": self.request_timeout,
            "max_tokens": self.max_tokens,
            "stream": self.streaming,
            "n": self.n,
            "temperature": self.temperature,
            **self.model_kwargs,
        }

    def completion_with_retry(
        self, run_manager: Optional[CallbackManagerForLLMRun] = None, **kwargs: Any
    ) -> Any:
        """Use tenacity to retry the completion call."""
        retry_decorator = _create_retry_decorator(self, run_manager=run_manager)

        @retry_decorator
        def _completion_with_retry(**kwargs: Any) -> Any:
            return self.client.create(**kwargs)

        return _completion_with_retry(**kwargs)

    def _combine_llm_outputs(self, llm_outputs: List[Optional[dict]]) -> dict:
        overall_token_usage: dict = {}
        for output in llm_outputs:
            if output is None:
                # Happens in streaming
                continue
            token_usage = output["token_usage"]
            for k, v in token_usage.items():
                if k in overall_token_usage:
                    overall_token_usage[k] += v
                else:
                    overall_token_usage[k] = v
        return {"token_usage": overall_token_usage, "model_name": self.model_name}
        
    def _stream(
        self,
        messages: List[BaseMessage],
        stop: Optional[List[str]] = None,
        run_manager: Optional[CallbackManagerForLLMRun] = None,
        **kwargs: Any,
    ) -> Iterator[ChatGenerationChunk]:
        message_dicts, params = self._create_message_dicts(messages, stop)
        params = {**params, **kwargs, "stream": True}

        default_chunk_class = AIMessageChunk
        for chunk in self.completion_with_retry(
            messages=message_dicts, run_manager=run_manager, **params
        ):
            if len(chunk["choices"]) == 0:
                continue
            choice = chunk["choices"][0]
            chunk = _convert_delta_to_message_chunk(
                choice["delta"], default_chunk_class
            )
            finish_reason = choice.get("finish_reason")
            generation_info = (
                dict(finish_reason=finish_reason) if finish_reason is not None else None
            )
            default_chunk_class = chunk.__class__
            yield ChatGenerationChunk(message=chunk, generation_info=generation_info)
            if run_manager:
                run_manager.on_llm_new_token(chunk.content, chunk=chunk)

    def _generate(
        self,
        messages: List[BaseMessage],
        stop: Optional[List[str]] = None,
        run_manager: Optional[CallbackManagerForLLMRun] = None,
        stream: Optional[bool] = None,
        **kwargs: Any,
    ) -> ChatResult:
        if stream if stream is not None else self.streaming:
            generation: Optional[ChatGenerationChunk] = None
            for chunk in self._stream(
                messages=messages, stop=stop, run_manager=run_manager, **kwargs
            ):
                if generation is None:
                    generation = chunk
                else:
                    generation += chunk
            assert generation is not None
            return ChatResult(generations=[generation])

        message_dicts, params = self._create_message_dicts(messages, stop)
        params = {**params, **kwargs}
        response = self.completion_with_retry(
            messages=message_dicts, run_manager=run_manager, **params
        )
        return self._create_chat_result(response)

    def _create_message_dicts(
        self, messages: List[BaseMessage], stop: Optional[List[str]]
    ) -> Tuple[List[Dict[str, Any]], Dict[str, Any]]:
        params = self._client_params
        if stop is not None:
            if "stop" in params:
                raise ValueError("`stop` found in both the input and default params.")
            params["stop"] = stop
        message_dicts = [convert_message_to_dict(m) for m in messages]
        return message_dicts, params

    def _create_chat_result(self, response: Mapping[str, Any]) -> ChatResult:
        generations = []
        for res in response["choices"]:
            message = convert_dict_to_message(res["message"])
            gen = ChatGeneration(
                message=message,
                generation_info=dict(finish_reason=res.get("finish_reason")),
            )
            generations.append(gen)
        token_usage = response.get("usage", {})
        llm_output = {"token_usage": token_usage, "model_name": self.model_name}
        return ChatResult(generations=generations, llm_output=llm_output)

    async def _astream(
        self,
        messages: List[BaseMessage],
        stop: Optional[List[str]] = None,
        run_manager: Optional[AsyncCallbackManagerForLLMRun] = None,
        **kwargs: Any,
    ) -> AsyncIterator[ChatGenerationChunk]:
        message_dicts, params = self._create_message_dicts(messages, stop)
        params = {**params, **kwargs, "stream": True}

        default_chunk_class = AIMessageChunk
        async for chunk in await acompletion_with_retry(
            self, messages=message_dicts, run_manager=run_manager, **params
        ):
            if len(chunk["choices"]) == 0:
                continue
            choice = chunk["choices"][0]
            chunk = _convert_delta_to_message_chunk(
                choice["delta"], default_chunk_class
            )
            finish_reason = choice.get("finish_reason")
            generation_info = (
                dict(finish_reason=finish_reason) if finish_reason is not None else None
            )
            default_chunk_class = chunk.__class__
            yield ChatGenerationChunk(message=chunk, generation_info=generation_info)
            if run_manager:
                await run_manager.on_llm_new_token(token=chunk.content, chunk=chunk)

    async def _agenerate(
        self,
        messages: List[BaseMessage],
        stop: Optional[List[str]] = None,
        run_manager: Optional[AsyncCallbackManagerForLLMRun] = None,
        stream: Optional[bool] = None,
        **kwargs: Any,
    ) -> ChatResult:
        if stream if stream is not None else self.streaming:
            generation: Optional[ChatGenerationChunk] = None
            async for chunk in self._astream(
                messages=messages, stop=stop, run_manager=run_manager, **kwargs
            ):
                if generation is None:
                    generation = chunk
                else:
                    generation += chunk
            assert generation is not None
            return ChatResult(generations=[generation])

        message_dicts, params = self._create_message_dicts(messages, stop)
        params = {**params, **kwargs}
        response = await acompletion_with_retry(
            self, messages=message_dicts, run_manager=run_manager, **params
        )
        return self._create_chat_result(response)

    @property
    def _identifying_params(self) -> Dict[str, Any]:
        """Get the identifying parameters."""
        return {**{"model_name": self.model_name}, **self._default_params}

    @property
    def _client_params(self) -> Dict[str, Any]:
        """Get the parameters used for the openai client."""
        openai_creds: Dict[str, Any] = {
            "openai_api_client_id": self.openai_api_client_id,
            "openai_api_client_secret": self.openai_api_client_secret,
            "openai_api_tokenurl": self.openai_api_tokenurl,
            "openai_api_url": self.openai_api_url,
            "model": self.model_name,
        }
        if self.openai_proxy:
            import openai

            openai.proxy = {"http": self.openai_proxy, "https": self.openai_proxy}  # type: ignore[assignment]  # noqa: E501
        return {**self._default_params, **openai_creds}

    def _get_invocation_params(
        self, stop: Optional[List[str]] = None, **kwargs: Any
    ) -> Dict[str, Any]:
        """Get the parameters used to invoke the model."""
        return {
            "model": self.model_name,
            **super()._get_invocation_params(stop=stop),
            **self._default_params,
            **kwargs,
        }  

    @property
    def _llm_type(self) -> str:
        """Return type of chat model."""
        return "sapopenai-chat"

    def _get_encoding_model(self) -> Tuple[str, tiktoken.Encoding]:
        tiktoken_ = _import_tiktoken()
        if self.tiktoken_model_name is not None:
            model = self.tiktoken_model_name
        else:
            model = self.model_name
            if model == "gpt-3.5-turbo":
                # gpt-3.5-turbo may change over time.
                # Returning num tokens assuming gpt-3.5-turbo-0301.
                model = "gpt-3.5-turbo-0301"
            elif model == "gpt-4":
                # gpt-4 may change over time.
                # Returning num tokens assuming gpt-4-0314.
                model = "gpt-4-0314"
        # Returns the number of tokens used by a list of messages.
        try:
            encoding = tiktoken_.encoding_for_model(model)
        except KeyError:
            logger.warning("Warning: model not found. Using cl100k_base encoding.")
            model = "cl100k_base"
            encoding = tiktoken_.get_encoding(model)
        return model, encoding

    def get_token_ids(self, text: str) -> List[int]:
        """Get the tokens present in the text with tiktoken package."""
        # tiktoken NOT supported for Python 3.7 or below
        if sys.version_info[1] <= 7:
            return super().get_token_ids(text)
        _, encoding_model = self._get_encoding_model()
        return encoding_model.encode(text)

    def get_num_tokens_from_messages(self, messages: List[BaseMessage]) -> int:
        """Calculate num tokens for gpt-3.5-turbo and gpt-4 with tiktoken package.

        Official documentation: https://github.com/openai/openai-cookbook/blob/
        main/examples/How_to_format_inputs_to_ChatGPT_models.ipynb"""
        if sys.version_info[1] <= 7:
            return super().get_num_tokens_from_messages(messages)
        model, encoding = self._get_encoding_model()
        if model.startswith("gpt-3.5-turbo-0301"):
            # every message follows <im_start>{role/name}\n{content}<im_end>\n
            tokens_per_message = 4
            # if there's a name, the role is omitted
            tokens_per_name = -1
        elif model.startswith("gpt-3.5-turbo") or model.startswith("gpt-4"):
            tokens_per_message = 3
            tokens_per_name = 1
        else:
            raise NotImplementedError(
                f"get_num_tokens_from_messages() is not presently implemented "
                f"for model {model}."
                "See https://github.com/openai/openai-python/blob/main/chatml.md for "
                "information on how messages are converted to tokens."
            )
        num_tokens = 0
        messages_dict = [convert_message_to_dict(m) for m in messages]
        for message in messages_dict:
            num_tokens += tokens_per_message
            for key, value in message.items():
                # Cast str(value) in case the message value is not a string
                # This occurs with function messages
                num_tokens += len(encoding.encode(str(value)))
                if key == "name":
                    num_tokens += tokens_per_name
        # every reply is primed with <im_start>assistant
        num_tokens += 3
        return num_tokens


import unittest

class TestProxyOpenAIStream(unittest.TestCase):

    def test_openai_stream(self):
        try:
            from saplangchainproxy.llm import SAPAzureOpenAI
            from langchain.callbacks.streaming_stdout import StreamingStdOutCallbackHandler
        except Exception as e:
            print(f"Error importing: {e}")
            self.assertTrue(False)

        llm = SAPAzureOpenAI(streaming=True, callbacks=[StreamingStdOutCallbackHandler()], temperature=0.0)
        resp = llm("Write me a song about sparkling water.")
        print(resp)

unittest.main()


import unittest

class TestProxyOpenAI(unittest.TestCase):
    def test_openai_chat(self):
        try:
            from saplangchainproxy.chat import SAPChatOpenAI
        except Exception as e:
            print(f"Error importing: {e}")
            self.assertTrue(False)
        chat = SAPChatOpenAI(model='gpt-4', temperature=0.0)
        from langchain.chains import ConversationChain
        from langchain.memory import ConversationBufferMemory

        conversation = ConversationChain(
            llm=chat,
            memory=ConversationBufferMemory()
        )
        test_string = "Hello I'm an AI."
        test_reply = conversation.run(f"Just answer with '{test_string}'. Nothing else.")
        self.assertTrue(test_string in test_reply)

    def test_openai_llm(self):
        try:
            from saplangchainproxy.llm import SAPAzureOpenAI
        except Exception as e:
            print(f"Error importing: {e}")
            self.assertTrue(False)
        llm = SAPAzureOpenAI(temperature=0.0)
        test_string = "Hello I'm an AI."
        test_reply = llm(f"Just answer with '{test_string}'. Nothing else.")
        self.assertTrue(test_string in test_reply)


# SAP Azure OpenAI communication handler class

# Completions
# deployment_id	                        description
# text-davinci-003 (default)	        Azure OpenAI text-davinci-003 model
# code-davinci-002	                    Azure OpenAI codex model
# alephalpha	                        Aleph Alpha luminous-base model
# gpt-35-turbo	                        Azure OpenAI ChatGPT (gpt-35-turbo) model
# gpt-4	                                Azure OpenAI GPT-4 model accepting 8k maximum tokens
# gpt-4-32k	                            Azure OpenAI GPT-4 model accepting 32k maximum tokens
# bloom-7b1	BLOOM                       model hosted on AI Core supporting more parameters
# gptj-full	GPT-J                       model hosted on AI Core supporting more parameters
# gcp-text-bison-001 (Preview)	        Google text-bison@001 model for text completions

# Embeddings
# deployment_id	                        description
# text-embedding-ada-002-v2	            Azure OpenAI text-embedding-ada-002 (Version 2) model

# Two endpoints are available:
# Completions:          POST <url>/api/v1/completions
# Embeddings:           POST <url>/api/v1/embeddings


from datetime import datetime
from datetime import timedelta

from oauthlib.oauth2 import BackendApplicationClient
from requests_oauthlib import OAuth2Session
import requests
import json

# OAuth2 token
token = {
            "sapopenai": {
                "envParams": {
                    "token": "OPENAI_TOKENURL",
                    "id": "OPENAI_CLIENTID",
                    "sec": "OPENAI_CLIENTSECRET"
                },
                "token": {}
                }
        }

class Completion:
    def __init__(self):
        Completion.openai_api_client_id = ""
        Completion.openai_api_client_secret = ""
        Completion.openai_api_url = ""
        Completion.openai_api_tokenurl = ""

    def getToken(service: str) -> str:
        global token
        # -------------- Env. Variables --------------->>>
        client = BackendApplicationClient(client_id=Completion.openai_api_client_id)
        # create an OAuth2 session
        oauth = OAuth2Session(client=client)
        if token[service]['token'] == {}:
            token[service]['token'] = oauth.fetch_token(token_url=Completion.openai_api_tokenurl, client_id=Completion.openai_api_client_id, client_secret=Completion.openai_api_client_secret)    
        elif datetime.fromtimestamp(token[service]['token']['expires_at']) - datetime.now() < timedelta(seconds=60):
            token[service]['token'] = oauth.fetch_token(token_url=Completion.openai_api_tokenurl, client_id=Completion.openai_api_client_id, client_secret=Completion.openai_api_client_secret)  
        return f"{token[service]['token']['token_type']} {token[service]['token']['access_token']}"
    
    def create(**kwargs):
        Completion.openai_api_client_id = kwargs['openai_api_client_id']
        Completion.openai_api_client_secret = kwargs['openai_api_client_secret']
        Completion.openai_api_tokenurl = kwargs['openai_api_tokenurl']
        Completion.openai_api_url = kwargs['openai_api_url']
        openAiUrl = Completion.openai_api_url + '/completions'

        headers = {
            "Content-Type": "application/json",
            "Authorization": Completion.getToken("sapopenai")
        }
        completionData = {}
        for key in kwargs:
            if key in ["request_timeout", "api_type", "api_version", "openai_api_client_id",
                       "openai_api_client_secret", "openai_api_tokenurl", "openai_api_url"]:
                continue
            if key == "engine": # Needs to be mapped to deployment_id for SAP
                engine = kwargs[key]
                if engine == "":
                    engine = "text-davinci-003"                    
                completionData['deployment_id'] = engine
            else:
                completionData[key] = kwargs[key]
        jCompletionData = json.dumps(completionData)
        response = requests.post(openAiUrl, headers=headers, data=jCompletionData)
        openAiAnswer = response.json()
        return openAiAnswer

class Embedding:
    def __init__(self):
        Completion.openai_api_client_id = ""
        Completion.openai_api_client_secret = ""
        Completion.openai_api_url = ""
        Completion.openai_api_tokenurl = ""

    def getToken(service: str) -> str:
        global token
        client = BackendApplicationClient(client_id=Completion.openai_api_client_id)
        # create an OAuth2 session
        oauth = OAuth2Session(client=client)
        if token[service]['token'] == {}:
            token[service]['token'] = oauth.fetch_token(token_url=Completion.openai_api_tokenurl, client_id=Completion.openai_api_client_id, client_secret=Completion.openai_api_client_secret)    
        elif datetime.fromtimestamp(token[service]['token']['expires_at']) - datetime.now() < timedelta(seconds=60):
            token[service]['token'] = oauth.fetch_token(token_url=Completion.openai_api_tokenurl, client_id=Completion.openai_api_client_id, client_secret=Completion.openai_api_client_secret)  
        return f"{token[service]['token']['token_type']} {token[service]['token']['access_token']}"
    
    def create(**kwargs):
        Completion.openai_api_client_id = kwargs['api_client_id']
        Completion.openai_api_client_secret = kwargs['api_client_secret']
        Completion.openai_api_tokenurl = kwargs['api_token_url']
        Completion.openai_api_url = kwargs['api_url']
        openAiUrl = Completion.openai_api_url + '/embeddings'
        headers = {
            "Content-Type": "application/json",
            "Authorization": Completion.getToken("sapopenai")
        }
        completionData = {}
        for key in kwargs:
            if key == "request_timeout":
                continue
            if key == "engine": # Needs to be mapped to deployment_id for SAP
                engine = kwargs[key]
                if engine == "" or engine == "text-embedding-ada-002":
                    engine = "text-embedding-ada-002-v2"                 
                completionData['deployment_id'] = engine
            else:
                completionData[key] = kwargs[key]
        jCompletionData = json.dumps(completionData)
        response = requests.post(openAiUrl, headers=headers, data=jCompletionData)
        openAiAnswer = response.json()
        return openAiAnswer
    
class ChatCompletion:
    def __init__(self):
        Completion.openai_api_client_id = ""
        Completion.openai_api_client_secret = ""
        Completion.openai_api_url = ""
        Completion.openai_api_tokenurl = ""

    def getToken(service: str) -> str:
        global token

        client = BackendApplicationClient(client_id=Completion.openai_api_client_id)
        # create an OAuth2 session
        oauth = OAuth2Session(client=client)
        if token[service]['token'] == {}:
            token[service]['token'] = oauth.fetch_token(token_url=Completion.openai_api_tokenurl, client_id=Completion.openai_api_client_id, client_secret=Completion.openai_api_client_secret)    
        elif datetime.fromtimestamp(token[service]['token']['expires_at']) - datetime.now() < timedelta(seconds=60):
            token[service]['token'] = oauth.fetch_token(token_url=Completion.openai_api_tokenurl, client_id=Completion.openai_api_client_id, client_secret=Completion.openai_api_client_secret)  
        return f"{token[service]['token']['token_type']} {token[service]['token']['access_token']}"
    
    def create(**kwargs):
        Completion.openai_api_client_id = kwargs['openai_api_client_id']
        Completion.openai_api_client_secret = kwargs['openai_api_client_secret']
        Completion.openai_api_tokenurl = kwargs['openai_api_tokenurl']
        Completion.openai_api_url = kwargs['openai_api_url']
        openAiUrl = Completion.openai_api_url + '/completions'
        headers = {
            "Content-Type": "application/json",
            "Authorization": Completion.getToken("sapopenai")
        }
        completionData = {}
        for key in kwargs:
            if key in ["request_timeout", "api_key", "api_base", "organization", "max_tokens", "openai_api_client_id", "openai_api_client_secret", "openai_api_tokenurl", "openai_api_url"] :
                continue
            if key == "model": # Needs to be mapped to deployment_id for SAP
                engine = kwargs[key]
                if engine == "gpt-3.5-turbo":
                    engine = "gpt-35-turbo"
                if engine == "":
                    engine = "gpt-35-turbo"                    
                completionData['deployment_id'] = engine
            else:
                completionData[key] = kwargs[key]
        jCompletionData = json.dumps(completionData)
        response = requests.post(openAiUrl, headers=headers, data=jCompletionData)
        openAiAnswer = response.json()
        return openAiAnswer